package org.apache.lucene.timeslots;

/**
 * In scoring, this query takes into account all spans and their payloads.
 * The source code comes from https://issues.apache.org/jira/browse/LUCENE-1341
 * slight modifications were made.
 */

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.lucene.index.IndexReader;
//import org.apache.lucene.index.Term;
import org.apache.lucene.index.TermPositions;
import org.apache.lucene.search.BooleanClause;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.Weight;
import org.apache.lucene.search.spans.SpanQuery;
import org.apache.lucene.search.spans.SpanNearQuery;
import org.apache.lucene.search.spans.SpanWeight;
import org.apache.lucene.search.spans.Spans;
import org.apache.lucene.search.spans.SpanScorer;
import org.apache.lucene.search.spans.TermSpans;
import org.apache.lucene.search.spans.NearSpansOrdered;
import org.apache.lucene.search.spans.NearSpansUnordered;

public class BoostingNearQuery extends SpanNearQuery {
	private static final long serialVersionUID = 12L;
	protected float docPayloadScore;
	private int docPayloadsSeen;
	boolean debug = false;

	private void DBG(String msg) { if(debug) { System.out.println(msg);} }

	public BoostingNearQuery(SpanQuery[] clauses, int slop, boolean inOrder) {
		super(clauses, slop, inOrder);
		DBG("BoostingNearQuery::constructor()");
	}

	public Weight createWeight(Searcher searcher) throws IOException {
		return new BoostingSpanWeight(this, searcher);
	}	

	public class BoostingSpanWeight extends SpanWeight {
		private static final long serialVersionUID = 13L;
		public BoostingSpanWeight(SpanQuery query, Searcher searcher) throws IOException {
			super (query, searcher);
			DBG("BoostingSpanWeight::constructor()");
		}

		@Override
			public Scorer scorer(IndexReader reader, boolean scoreDocsInOrder,
					boolean topScorer) throws IOException {
				return new BoostingSpanScorer(query.getSpans(reader), this,
						similarity, reader.norms(query.getField()));
			}	
	}

	public class BoostingSpanScorer extends SpanScorer 
	{
		private static final long serialVersionUID = 14L;
		//TODO: is this the best way to allocate this?
		byte[] payload = new byte[256];
		Spans[] subSpans=null;
		private ArrayList<TermPositions> termsPositions = new ArrayList<TermPositions>();
		Similarity similarity = getSimilarity();

		protected BoostingSpanScorer(Spans spans, Weight weight, Similarity similarity, byte[] norms) throws IOException {
			super (spans, weight, similarity, norms);
			DBG("BoostingSpanScorer::constructor()");
			Spans[] spansArr = new Spans[1];
			spansArr[0] = spans;
			getTermsPositions(spansArr, termsPositions);
		}
		// Get the TermPositions associated with all underlying subspans
		public void getTermsPositions(Spans[] subSpans, ArrayList<TermPositions> termsPositions) 
		{
			DBG("BoostingSpanScorer::getTermsPositions()");
			for (int i=0;i<subSpans.length;i++) {
				if (subSpans[i] instanceof TermSpans) {
					DBG("  termsPositions.add() doc="+(subSpans[i]).doc()+" start="+(subSpans[i]).start()+" end="+(subSpans[i]).end());
					termsPositions.add(((TermSpans)subSpans[i]).getPositions());
				} else if (subSpans[i] instanceof NearSpansOrdered) {
					DBG("  NearSpansOrdered");
					getTermsPositions(((NearSpansOrdered)subSpans[i]).getSubSpans(), termsPositions);
				} else if (subSpans[i] instanceof NearSpansUnordered) {
					DBG("  NearSpansUnrdered");
					getTermsPositions(((NearSpansUnordered)subSpans[i]).getSubSpans(), termsPositions);
				}
			}
		}

		@Override
		protected boolean setFreqCurrentDoc() throws IOException 
		{
			if (!more) {
				return false;
			}
			doc = spans.doc();
			DBG("BoostingSpanScorer::setFreqCurrentDoc() doc="+doc);
			freq = 0.0f;
			docPayloadScore = 0;
			docPayloadsSeen = 0;
			while (more && doc == spans.doc()) {
				float spanScore = 1; // maximum possible score
				DBG("termsPositions.size(): "+termsPositions.size());
				for (Iterator<TermPositions> it = termsPositions.iterator(); it.hasNext();) {
					DBG("  next positions");
					TermPositions positions = it.next();
					if(positions.isPayloadAvailable()) {
						payload = positions.getPayload(payload, 0);
						float payloadScore = similarity.scorePayload(doc, field, 0, 0, payload, 0, positions.getPayloadLength());
						if (spanScore > payloadScore) { 
							spanScore = payloadScore;
						}
						DBG("    payload: " + similarity.scorePayload(doc, field, 0, 0, payload, 0, positions.getPayloadLength()));

						//payloadScore += similarity.scorePayload(doc, field, 0, 0, payload, 0, positions.getPayloadLength());
					}
				}
				docPayloadScore += spanScore;
				docPayloadsSeen++;
				int matchLength = spans.end() - spans.start();
				DBG("  matchLength="+matchLength+" spanScore="+spanScore);
				freq += getSimilarity().sloppyFreq(matchLength);
				more = spans.next();
			}
			return more || (freq != 0);
		}		 

		@Override
		public float score() throws IOException 
		{
			DBG("BoostingSpanScorer::score() = "+super.score()+" * "+docPayloadScore+" / "+docPayloadsSeen);
			return super.score() * (docPayloadsSeen > 0 ? (docPayloadScore / docPayloadsSeen) : 1);
		}		 

		@Override
		public Explanation explain(int doc) throws IOException 
		{
			Explanation result = new Explanation();
			Explanation nonPayloadExpl = super.explain(doc);
			result.addDetail(nonPayloadExpl);
			Explanation payloadBoost = new Explanation();
			result.addDetail(payloadBoost);
			float avgPayloadScore =  (docPayloadsSeen > 0 ? (docPayloadScore / docPayloadsSeen) : 1); 
			payloadBoost.setValue(avgPayloadScore);
			payloadBoost.setDescription("scorePayload(...)  score / seen = "+docPayloadScore+" / "+docPayloadsSeen);
			result.setValue(nonPayloadExpl.getValue() * avgPayloadScore);
			result.setDescription("btq, product of:");
			return result;
		}
	}



}

