package org.apache.lucene.timeslots;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.TermPositions;
import org.apache.lucene.search.Explanation;
import org.apache.lucene.search.Scorer;
import org.apache.lucene.search.Searcher;
import org.apache.lucene.search.Similarity;
import org.apache.lucene.search.Weight;
import org.apache.lucene.search.spans.NearSpansOrdered;
import org.apache.lucene.search.spans.NearSpansUnordered;
import org.apache.lucene.search.spans.SpanNearQuery;
import org.apache.lucene.search.spans.SpanQuery;
import org.apache.lucene.search.spans.SpanScorer;
import org.apache.lucene.search.spans.SpanWeight;
import org.apache.lucene.search.spans.Spans;
import org.apache.lucene.search.payloads.PayloadFunction;
import org.apache.lucene.search.payloads.AveragePayloadFunction;
import org.apache.lucene.util.ToStringUtils;

import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.ArrayList;

public class TSNearQuery extends SpanNearQuery {
	private static final long serialVersionUID = 9L;
	protected String fieldName;
	protected PayloadFunction function;

	public TSNearQuery(SpanQuery[] clauses, int slop, boolean inOrder) {
		this(clauses, slop, inOrder, new AveragePayloadFunction());
		System.out.println("TSNearQuery::constructor()");
	}

	public TSNearQuery(SpanQuery[] clauses, int slop, boolean inOrder,
			PayloadFunction function) {
		super(clauses, slop, inOrder);
		System.out.println("TSNearQuery::constructor()");
		fieldName = clauses[0].getField(); // all clauses must have same field
		this.function = function;
	}

	@Override
		public Weight createWeight(Searcher searcher) throws IOException {
			return new TSNearSpanWeight(this, searcher);
		}

	@Override
		public Object clone() {
			int sz = clauses.size();
			SpanQuery[] newClauses = new SpanQuery[sz];

			for (int i = 0; i < sz; i++) {
				newClauses[i] = (SpanQuery) clauses.get(i).clone();
			}
			TSNearQuery boostingNearQuery = new TSNearQuery(newClauses, slop,
					inOrder);
			boostingNearQuery.setBoost(getBoost());
			return boostingNearQuery;
		}

	@Override
		public String toString(String field) {
			StringBuilder buffer = new StringBuilder();
			buffer.append("payloadNear([");
			Iterator<SpanQuery> i = clauses.iterator();
			while (i.hasNext()) {
				SpanQuery clause = i.next();
				buffer.append(clause.toString(field));
				if (i.hasNext()) {
					buffer.append(", ");
				}
			}
			buffer.append("], ");
			buffer.append(slop);
			buffer.append(", ");
			buffer.append(inOrder);
			buffer.append(")");
			buffer.append(ToStringUtils.boost(getBoost()));
			return buffer.toString();
		}

	@Override
		public int hashCode() {
			final int prime = 31;
			int result = super.hashCode();
			result = prime * result + ((fieldName == null) ? 0 : fieldName.hashCode());
			result = prime * result + ((function == null) ? 0 : function.hashCode());
			return result;
		}

	@Override
		public boolean equals(Object obj) {
			if (this == obj)
				return true;
			if (!super.equals(obj))
				return false;
			if (getClass() != obj.getClass())
				return false;
			TSNearQuery other = (TSNearQuery) obj;
			if (fieldName == null) {
				if (other.fieldName != null)
					return false;
			} else if (!fieldName.equals(other.fieldName))
				return false;
			if (function == null) {
				if (other.function != null)
					return false;
			} else if (!function.equals(other.function))
				return false;
			return true;
		}

	public class TSNearSpanWeight extends SpanWeight {
		private static final long serialVersionUID = 10L;
		public TSNearSpanWeight(SpanQuery query, Searcher searcher)
			throws IOException {
			super(query, searcher);
			System.out.println("TSNearSpanWeight::constructor()");
		}

		@Override
			public Scorer scorer(IndexReader reader, boolean scoreDocsInOrder,
					boolean topScorer) throws IOException {
				return new TSNearSpanScorer(query.getSpans(reader), this,
						similarity, reader.norms(query.getField()));
			}
	}

	public class TSNearSpanScorer extends SpanScorer {
		private static final long serialVersionUID = 11L;
		Spans spans;

		protected float payloadScore;
		private int payloadsSeen;
		Similarity similarity = getSimilarity();

		protected TSNearSpanScorer(Spans spans, Weight weight,
				Similarity similarity, byte[] norms) throws IOException {
			super(spans, weight, similarity, norms);
			System.out.println("TSNearSpanScorer::constructor()");
			this.spans = spans;
//			for (Span span : spans) {
//				System.out.println("span doc="+span.doc);
//			}
		}

		// Get the payloads associated with all underlying subspans
		public void getPayloads(Spans[] subSpans) throws IOException {
			System.out.println("getPayloads() subSpans.length="+subSpans.length);
			for (int i = 0; i < subSpans.length; i++) {
				if (subSpans[i] instanceof NearSpansOrdered) {
					if (((NearSpansOrdered) subSpans[i]).isPayloadAvailable()) {
						processPayloads(((NearSpansOrdered) subSpans[i]).getPayload(),
								subSpans[i].start(), subSpans[i].end());
					}
					getPayloads(((NearSpansOrdered) subSpans[i]).getSubSpans());
				} else if (subSpans[i] instanceof NearSpansUnordered) {
					if (((NearSpansUnordered) subSpans[i]).isPayloadAvailable()) {
						processPayloads(((NearSpansUnordered) subSpans[i]).getPayload(),
								subSpans[i].start(), subSpans[i].end());
	  				}
					getPayloads(((NearSpansUnordered) subSpans[i]).getSubSpans());
				}
			}
		}

		/**
		 * By default, uses the {@link PayloadFunction} to score the payloads, but
		 * can be overridden to do other things.
		 * 
		 * @param payLoads The payloads
		 * @param start The start position of the span being scored
		 * @param end The end position of the span being scored
		 * 
		 * @see Spans
		 */
		protected void processPayloads(Collection<byte[]> payLoads, int start, int end) {
			System.out.println("processPayloads() payLoads.size() = "+payLoads.size());
			for (final byte[] thePayload : payLoads) {
				payloadScore = function.currentScore(doc, fieldName, start, end,
						payloadsSeen, payloadScore, similarity.scorePayload(doc, fieldName,
							spans.start(), spans.end(), thePayload, 0, thePayload.length));
				System.out.println("  payloadScore: "+payloadScore);
				++payloadsSeen;
			}
		}

		//
		@Override
			protected boolean setFreqCurrentDoc() throws IOException {
				if (!more) {
					return false;
				}
				Spans[] spansArr = new Spans[1];
				spansArr[0] = spans;
				payloadScore = 0;
				payloadsSeen = 0;
				getPayloads(spansArr);
				return super.setFreqCurrentDoc();
			}

		@Override
			public float score() throws IOException {

				return super.score()
					* function.docScore(doc, fieldName, payloadsSeen, payloadScore);
			}

		@Override
			protected Explanation explain(int doc) throws IOException {
				Explanation result = new Explanation();
				Explanation nonPayloadExpl = super.explain(doc);
				result.addDetail(nonPayloadExpl);
				Explanation payloadBoost = new Explanation();
				result.addDetail(payloadBoost);
				float avgPayloadScore = (payloadsSeen > 0 ? (payloadScore / payloadsSeen)
						: 1);
				payloadBoost.setValue(avgPayloadScore);
				payloadBoost.setDescription("scorePayload(...) payloadScore="+payloadScore+" / payloadsSeen="+payloadsSeen);
				result.setValue(nonPayloadExpl.getValue() * avgPayloadScore);
				result.setDescription("bnq, product of:");
				return result;
			}
	}

}

